@@ -27,10 +27,20 @@ QT_BEGIN_NAMESPACE
 
 Q_DECLARE_LOGGING_CATEGORY(lcQmlTypeCompiler);
 
+// This class primarily resolves component boundaries in a document.
+// With the information about boundaries, it then goes on to resolve aliases and generalized
+// group properties. Both rely on IDs as first part of their expressions and the IDs have
+// to be located in surrounding components. That's why we have to do this with the component
+// boundaries in mind.
+
+class QQmlComponentAndAliasResolverBase
+{
+    Q_DECLARE_TR_FUNCTIONS(QQmlComponentAndAliasResolverBase)
+};
+
 template<typename ObjectContainer>
-class QQmlComponentAndAliasResolver
+class QQmlComponentAndAliasResolver : public QQmlComponentAndAliasResolverBase
 {
-    Q_DECLARE_TR_FUNCTIONS(QQmlComponentAndAliasResolver)
 public:
     using CompiledObject = typename ObjectContainer::CompiledObject;
     using CompiledBinding = typename ObjectContainer::CompiledBinding;
@@ -53,13 +63,17 @@ private:
     void allocateNamedObjects(CompiledObject *object) const;
     void setObjectId(int index) const;
     [[nodiscard]] bool markAsComponent(int index) const;
-    [[nodiscard]] AliasResolutionResult resolveAliasesInObject(int objectIndex, QQmlError *error);
+    [[nodiscard]] AliasResolutionResult resolveAliasesInObject(
+            const CompiledObject &component, int objectIndex, QQmlError *error);
+    void resolveGeneralizedGroupProperty(const CompiledObject &component, CompiledBinding *binding);
     [[nodiscard]] bool wrapImplicitComponent(CompiledBinding *binding);
 
     [[nodiscard]] QQmlError findAndRegisterImplicitComponents(
             const CompiledObject *obj, const QQmlPropertyCache::ConstPtr &propertyCache);
     [[nodiscard]] QQmlError collectIdsAndAliases(int objectIndex);
     [[nodiscard]] QQmlError resolveAliases(int componentIndex);
+    void resolveGeneralizedGroupProperties(int componentIndex);
+    [[nodiscard]] QQmlError resolveComponentsInInlineComponentRoot(int root);
 
     QString stringAt(int idx) const { return m_compiler->stringAt(idx); }
     QV4::ResolvedTypeReference *resolvedType(int id) const { return m_compiler->resolvedType(id); }
@@ -109,6 +123,7 @@ private:
     // indices of the objects that are actually Component {}
     QVector<quint32> m_componentRoots;
     QVector<int> m_objectsWithAliases;
+    QVector<CompiledBinding *> m_generalizedGroupProperties;
     typename ObjectContainer::IdToObjectMap m_idToObjectIndex;
 };
 
@@ -180,12 +195,42 @@ QQmlError QQmlComponentAndAliasResolver<ObjectContainer>::findAndRegisterImplici
             continue;
 
         if (!wrapImplicitComponent(binding))
-            return error(binding, tr("Cannot wrap implicit component"));
+            return error(binding, QQmlComponentAndAliasResolverBase::tr("Cannot wrap implicit component"));
     }
 
     return QQmlError();
 }
 
+template<typename ObjectContainer>
+QQmlError QQmlComponentAndAliasResolver<ObjectContainer>::resolveComponentsInInlineComponentRoot(
+        int root)
+{
+    // Find implicit components in the inline component itself. Also warn about inline
+    // components being explicit components.
+
+    const auto rootObj = m_compiler->objectAt(root);
+    Q_ASSERT(rootObj->hasFlag(QV4::CompiledData::Object::IsInlineComponentRoot));
+
+    if (const int typeName = rootObj->inheritedTypeNameIndex) {
+        const auto *tref = resolvedType(typeName);
+        Q_ASSERT(tref);
+        if (tref->type().metaObject() == &QQmlComponent::staticMetaObject) {
+            qCWarning(lcQmlTypeCompiler).nospace().noquote()
+                    << m_compiler->url().toString() << ":" << rootObj->location.line() << ":"
+                    << rootObj->location.column()
+                    << ": Using a Component as the root of an inline component is deprecated: "
+                       "inline components are "
+                       "automatically wrapped into Components when needed.";
+            return QQmlError();
+        }
+    }
+
+    const QQmlPropertyCache::ConstPtr rootCache = m_propertyCaches->at(root);
+    Q_ASSERT(rootCache);
+
+    return findAndRegisterImplicitComponents(rootObj, rootCache);
+}
+
 // Resolve ignores everything relating to inline components, except for implicit components.
 template<typename ObjectContainer>
 QQmlError QQmlComponentAndAliasResolver<ObjectContainer>::resolve(int root)
@@ -196,6 +241,12 @@ QQmlError QQmlComponentAndAliasResolver<ObjectContainer>::resolve(int root)
     // on the left hand side is of QQmlComponent type.
     const int objCountWithoutSynthesizedComponents = m_compiler->objectCount();
 
+    if (root != 0) {
+        const QQmlError error = resolveComponentsInInlineComponentRoot(root);
+        if (error.isValid())
+            return error;
+    }
+
     // root+1, as ic root is handled at the end
     const int startObjectIndex = root == 0 ? root : root+1;
 
@@ -207,6 +258,21 @@ QQmlError QQmlComponentAndAliasResolver<ObjectContainer>::resolve(int root)
                 = obj->hasFlag(QV4::CompiledData::Object::IsPartOfInlineComponent);
         QQmlPropertyCache::ConstPtr cache = m_propertyCaches->at(i);
 
+        if (root == 0) {
+            // normal component root, skip over anything inline component related
+            if (isInlineComponentRoot || isPartOfInlineComponent)
+                continue;
+        } else if (!isPartOfInlineComponent || isInlineComponentRoot) {
+            // When handling an inline component, stop where the inline component ends
+            // Note: We do not support nested inline components. Therefore, isInlineComponentRoot
+            //       tells us that the element after the current inline component is again an
+            //       inline component
+            break;
+        }
+
+        if (obj->inheritedTypeNameIndex == 0 && !cache)
+            continue;
+
         bool isExplicitComponent = false;
         if (obj->inheritedTypeNameIndex) {
             auto *tref = resolvedType(obj->inheritedTypeNameIndex);
@@ -215,33 +281,6 @@ QQmlError QQmlComponentAndAliasResolver<ObjectContainer>::resolve(int root)
                 isExplicitComponent = true;
         }
 
-        if (isInlineComponentRoot && isExplicitComponent) {
-            qCWarning(lcQmlTypeCompiler).nospace().noquote()
-                    << m_compiler->url().toString() << ":" << obj->location.line() << ":"
-                    << obj->location.column()
-                    << ": Using a Component as the root of an inline component is deprecated: "
-                       "inline components are "
-                       "automatically wrapped into Components when needed.";
-        }
-
-        if (root == 0) {
-            // normal component root, skip over anything inline component related
-            if (isInlineComponentRoot || isPartOfInlineComponent)
-                continue;
-        } else if (!isPartOfInlineComponent || isInlineComponentRoot) {
-            // We've left the current inline component (potentially entered a new one),
-            // but we still need to resolve implicit components which are part of inline components.
-            if (cache && !isExplicitComponent) {
-                const QQmlError error = findAndRegisterImplicitComponents(obj, cache);
-                if (error.isValid())
-                    return error;
-            }
-            break;
-        }
-
-        if (obj->inheritedTypeNameIndex == 0 && !cache)
-            continue;
-
         if (!isExplicitComponent) {
             if (cache) {
                 const QQmlError error = findAndRegisterImplicitComponents(obj, cache);
@@ -252,7 +291,7 @@ QQmlError QQmlComponentAndAliasResolver<ObjectContainer>::resolve(int root)
         }
 
         if (!markAsComponent(i))
-            return error(obj, tr("Cannot mark object as component"));
+            return error(obj, QQmlComponentAndAliasResolverBase::tr("Cannot mark object as component"));
 
         // check if this object is the root
         if (i == 0) {
@@ -266,14 +305,14 @@ QQmlError QQmlComponentAndAliasResolver<ObjectContainer>::resolve(int root)
         }
 
         if (obj->functionCount() > 0)
-            return error(obj, tr("Component objects cannot declare new functions."));
+            return error(obj, QQmlComponentAndAliasResolverBase::tr("Component objects cannot declare new functions."));
         if (obj->propertyCount() > 0 || obj->aliasCount() > 0)
-            return error(obj, tr("Component objects cannot declare new properties."));
+            return error(obj, QQmlComponentAndAliasResolverBase::tr("Component objects cannot declare new properties."));
         if (obj->signalCount() > 0)
-            return error(obj, tr("Component objects cannot declare new signals."));
+            return error(obj, QQmlComponentAndAliasResolverBase::tr("Component objects cannot declare new signals."));
 
         if (obj->bindingCount() == 0)
-            return error(obj, tr("Cannot create empty component specification"));
+            return error(obj, QQmlComponentAndAliasResolverBase::tr("Cannot create empty component specification"));
 
         const auto rootBinding = obj->bindingsBegin();
         const auto bindingsEnd = obj->bindingsEnd();
@@ -284,12 +323,12 @@ QQmlError QQmlComponentAndAliasResolver<ObjectContainer>::resolve(int root)
             if (b->propertyNameIndex == 0)
                 continue;
 
-            return error(b, tr("Component elements may not contain properties other than id"));
+            return error(b, QQmlComponentAndAliasResolverBase::tr("Component elements may not contain properties other than id"));
         }
 
         if (auto b = rootBinding;
                 b->type() != QV4::CompiledData::Binding::Type_Object || ++b != bindingsEnd) {
-            return error(obj, tr("Invalid component body specification"));
+            return error(obj, QQmlComponentAndAliasResolverBase::tr("Invalid component body specification"));
         }
 
         // For the root object, we are going to collect ids/aliases and resolve them for as a
@@ -304,6 +343,7 @@ QQmlError QQmlComponentAndAliasResolver<ObjectContainer>::resolve(int root)
 
         m_idToObjectIndex.clear();
         m_objectsWithAliases.clear();
+        m_generalizedGroupProperties.clear();
 
         if (const QQmlError error = collectIdsAndAliases(rootBinding->value.objectIndex);
                 error.isValid()) {
@@ -314,17 +354,24 @@ QQmlError QQmlComponentAndAliasResolver<ObjectContainer>::resolve(int root)
 
         if (const QQmlError error = resolveAliases(m_componentRoots.at(i)); error.isValid())
             return error;
+
+        resolveGeneralizedGroupProperties(m_componentRoots.at(i));
     }
 
     // Collect ids and aliases for root
     m_idToObjectIndex.clear();
     m_objectsWithAliases.clear();
+    m_generalizedGroupProperties.clear();
 
     if (const QQmlError error = collectIdsAndAliases(root); error.isValid())
         return error;
 
     allocateNamedObjects(m_compiler->objectAt(root));
-    return resolveAliases(root);
+    if (const QQmlError error = resolveAliases(root); error.isValid())
+        return error;
+
+    resolveGeneralizedGroupProperties(root);
+    return QQmlError();
 }
 
 template<typename ObjectContainer>
@@ -334,7 +381,7 @@ QQmlError QQmlComponentAndAliasResolver<ObjectContainer>::collectIdsAndAliases(i
 
     if (obj->idNameIndex != 0) {
         if (m_idToObjectIndex.contains(obj->idNameIndex))
-            return error(obj->locationOfIdProperty, tr("id is not unique"));
+            return error(obj->locationOfIdProperty, QQmlComponentAndAliasResolverBase::tr("id is not unique"));
         setObjectId(objectIndex);
         m_idToObjectIndex.insert(obj->idNameIndex, objectIndex);
     }
@@ -349,9 +396,19 @@ QQmlError QQmlComponentAndAliasResolver<ObjectContainer>::collectIdsAndAliases(i
     for (auto binding = obj->bindingsBegin(), end = obj->bindingsEnd();
          binding != end; ++binding) {
         switch (binding->type()) {
+        case QV4::CompiledData::Binding::Type_GroupProperty: {
+            const auto *inner = m_compiler->objectAt(binding->value.objectIndex);
+            if (m_compiler->stringAt(inner->inheritedTypeNameIndex).isEmpty()) {
+                const auto cache = m_propertyCaches->at(objectIndex);
+                if (!cache || !cache->property(
+                            m_compiler->stringAt(binding->propertyNameIndex), nullptr, nullptr)) {
+                    m_generalizedGroupProperties.append(binding);
+                }
+            }
+        }
+        Q_FALLTHROUGH();
         case QV4::CompiledData::Binding::Type_Object:
         case QV4::CompiledData::Binding::Type_AttachedProperty:
-        case QV4::CompiledData::Binding::Type_GroupProperty:
             if (const QQmlError error = collectIdsAndAliases(binding->value.objectIndex);
                     error.isValid()) {
                 return error;
@@ -381,13 +438,14 @@ QQmlError QQmlComponentAndAliasResolver<ObjectContainer>::resolveAliases(int com
         for (int objectIndex: std::as_const(m_objectsWithAliases)) {
 
             QQmlError error;
-            const auto result = resolveAliasesInObject(objectIndex, &error);
+            const auto &component = *m_compiler->objectAt(componentIndex);
+            const auto result = resolveAliasesInObject(component, objectIndex, &error);
             if (error.isValid())
                 return error;
 
             if (result == AllAliasesResolved) {
                 QQmlError error = aliasCacheCreator.appendAliasesToPropertyCache(
-                            *m_compiler->objectAt(componentIndex), objectIndex, m_enginePrivate);
+                            component, objectIndex, m_enginePrivate);
                 if (error.isValid())
                     return error;
                 atLeastOneAliasResolved = true;
@@ -405,13 +463,22 @@ QQmlError QQmlComponentAndAliasResolver<ObjectContainer>::resolveAliases(int com
         const CompiledObject *obj = m_compiler->objectAt(m_objectsWithAliases.first());
         for (auto alias = obj->aliasesBegin(), end = obj->aliasesEnd(); alias != end; ++alias) {
             if (!alias->hasFlag(QV4::CompiledData::Alias::Resolved))
-                return error(alias->location, tr("Circular alias reference detected"));
+                return error(alias->location, QQmlComponentAndAliasResolverBase::tr("Circular alias reference detected"));
         }
     }
 
     return QQmlError();
 }
 
+template<typename ObjectContainer>
+void QQmlComponentAndAliasResolver<ObjectContainer>::resolveGeneralizedGroupProperties(
+        int componentIndex)
+{
+    const auto &component = *m_compiler->objectAt(componentIndex);
+    for (CompiledBinding *binding : m_generalizedGroupProperties)
+        resolveGeneralizedGroupProperty(component, binding);
+}
+
 QT_END_NAMESPACE
 
 #endif // QQMLCOMPONENTANDALIASRESOLVER_P_H
