@@ -25,6 +25,7 @@
 #include <private/qv4calldata_p.h>
 
 #include <QtCore/qsharedpointer.h>
+#include <stack>
 
 QT_BEGIN_NAMESPACE
 
@@ -44,7 +45,7 @@ struct ControlFlow;
 struct ControlFlowCatch;
 struct ControlFlowFinally;
 
-class Q_QML_COMPILER_PRIVATE_EXPORT CodegenWarningInterface
+class Q_QML_COMPILER_EXPORT CodegenWarningInterface
 {
 public:
     virtual void reportVarUsedBeforeDeclaration(const QString &name, const QString &fileName,
@@ -59,7 +60,7 @@ inline CodegenWarningInterface *defaultCodegenWarningInterface()
     return &iface;
 }
 
-class Q_QML_COMPILER_PRIVATE_EXPORT Codegen: protected QQmlJS::AST::Visitor
+class Q_QML_COMPILER_EXPORT Codegen: protected QQmlJS::AST::Visitor
 {
 protected:
     using BytecodeGenerator = QV4::Moth::BytecodeGenerator;
@@ -69,18 +70,12 @@ public:
             CodegenWarningInterface *iface = defaultCodegenWarningInterface(),
             bool storeSourceLocations = false);
 
-    void generateFromProgram(const QString &fileName,
-                             const QString &finalUrl,
-                             const QString &sourceCode,
-                             QQmlJS::AST::Program *ast,
-                             Module *module,
-                             ContextType contextType = ContextType::Global);
+    void generateFromProgram(
+            const QString &sourceCode, QQmlJS::AST::Program *ast, Module *module,
+            ContextType contextType = ContextType::Global);
 
-    void generateFromModule(const QString &fileName,
-                            const QString &finalUrl,
-                            const QString &sourceCode,
-                            QQmlJS::AST::ESModule *ast,
-                            Module *module);
+    void generateFromModule(const QString &sourceCode, QQmlJS::AST::ESModule *ast, Module *module);
+    void generateFromModule(const Value &value, Module *module);
 
 public:
     class VolatileMemoryLocationScanner;
@@ -190,7 +185,9 @@ public:
             global(false),
             qmlGlobal(false),
             throwsReferenceError(false),
-            subscriptLoadedForCall(false)
+            subscriptLoadedForCall(false),
+            isOptional(false),
+            hasSavedCallBaseSlot(false)
         {}
 
         Reference(const Reference &) = default;
@@ -256,8 +253,8 @@ public:
         static Reference
         fromMember(const Reference &baseRef, const QString &name,
                    QQmlJS::SourceLocation sourceLocation = QQmlJS::SourceLocation(),
-                   Moth::BytecodeGenerator::Label jumpLabel = Moth::BytecodeGenerator::Label(),
-                   Moth::BytecodeGenerator::Label targetLabel = Moth::BytecodeGenerator::Label())
+                   bool isOptional = false,
+                   std::vector<Moth::BytecodeGenerator::Jump> *optionalChainJumpsToPatch = nullptr)
         {
             Q_ASSERT(baseRef.isValid());
             Reference r(baseRef.codegen, Member);
@@ -265,8 +262,8 @@ public:
             r.propertyNameIndex = r.codegen->registerString(name);
             r.requiresTDZCheck = baseRef.requiresTDZCheck;
             r.sourceLocation = sourceLocation;
-            r.optionalChainJumpLabel.reset(new Moth::BytecodeGenerator::Label(jumpLabel));
-            r.optionalChainTargetLabel.reset(new Moth::BytecodeGenerator::Label(targetLabel));
+            r.optionalChainJumpsToPatch = optionalChainJumpsToPatch;
+            r.isOptional = isOptional;
             return r;
         }
         static Reference fromSuperProperty(const Reference &property) {
@@ -276,14 +273,13 @@ public:
             r.subscriptRequiresTDZCheck = property.requiresTDZCheck;
             return r;
         }
-        static Reference fromSubscript(const Reference &baseRef, const Reference &subscript, Moth::BytecodeGenerator::Label targetLabel = Moth::BytecodeGenerator::Label()) {
+        static Reference fromSubscript(const Reference &baseRef, const Reference &subscript) {
             Q_ASSERT(baseRef.isStackSlot());
             Reference r(baseRef.codegen, Subscript);
             r.elementBase = baseRef.stackSlot();
             r.elementSubscript = subscript.asRValue();
             r.requiresTDZCheck = baseRef.requiresTDZCheck;
             r.subscriptRequiresTDZCheck = subscript.requiresTDZCheck;
-            r.optionalChainTargetLabel.reset(new Moth::BytecodeGenerator::Label(targetLabel));
             return r;
         }
         static Reference fromConst(Codegen *cg, QV4::ReturnedValue constant) {
@@ -373,9 +369,12 @@ public:
         quint32 qmlGlobal:1;
         quint32 throwsReferenceError:1;
         quint32 subscriptLoadedForCall:1;
+        quint32 isOptional: 1;
+        quint32 hasSavedCallBaseSlot: 1;
         QQmlJS::SourceLocation sourceLocation = QQmlJS::SourceLocation();
-        QSharedPointer<Moth::BytecodeGenerator::Label> optionalChainJumpLabel;
-        QSharedPointer<Moth::BytecodeGenerator::Label> optionalChainTargetLabel;
+        std::vector<Moth::BytecodeGenerator::Jump> *optionalChainJumpsToPatch = nullptr;
+        int savedCallBaseSlot = -1;
+        int savedCallPropertyNameIndex = -1;
 
     private:
         void storeAccumulator() const;
@@ -706,7 +705,8 @@ public:
     QQmlJS::DiagnosticMessage error() const;
     QUrl url() const;
 
-    Reference binopHelper(QSOperator::Op oper, Reference &left, Reference &right);
+    Reference binopHelper(QQmlJS::AST::BinaryExpression *ast, QSOperator::Op oper, Reference &left,
+                          Reference &right);
     Reference jumpBinop(QSOperator::Op oper, Reference &left, Reference &right);
     struct Arguments { int argc; int argv; bool hasSpread; };
     Arguments pushArgs(QQmlJS::AST::ArgumentList *args);
@@ -726,11 +726,15 @@ public:
             const QString &name, bool lhs,
             const QQmlJS::SourceLocation &accessLocation = QQmlJS::SourceLocation());
 
-    QV4::CompiledData::CompilationUnit generateCompilationUnit(bool generateUnitData = true);
-    static QV4::CompiledData::CompilationUnit compileModule(
+    QQmlRefPointer<QV4::CompiledData::CompilationUnit> generateCompilationUnit(
+            bool generateUnitData = true);
+    static QQmlRefPointer<QV4::CompiledData::CompilationUnit> compileModule(
             bool debugMode, const QString &url, const QString &sourceCode,
             const QDateTime &sourceTimeStamp, QList<QQmlJS::DiagnosticMessage> *diagnostics);
 
+    static const QV4::CompiledData::Unit *generateNativeModuleUnitData(
+            bool debugMode, const QString &url, const QV4::Value &value);
+
     Context *currentContext() const { return _context; }
     BytecodeGenerator *generator() const { return bytecodeGenerator; }
 
@@ -801,8 +805,15 @@ protected:
     bool _tailCallsAreAllowed = true;
     bool storeSourceLocations = false;
     QSet<QString> m_globalNames;
+
+    struct OptionalChainState
+    {
+        QQmlJS::AST::Node *tailNodeOfChain = nullptr;
+        std::vector<Moth::BytecodeGenerator::Jump> jumpsToPatch;
+        bool actuallyHasOptionals = false;
+    };
     QSet<QQmlJS::AST::Node*> m_seenOptionalChainNodes;
-    QHash<QQmlJS::AST::Node*, Moth::BytecodeGenerator::Label> m_optionalChainLabels;
+    std::stack<OptionalChainState> m_optionalChainsStates;
 
     ControlFlow *controlFlow = nullptr;
 
@@ -836,11 +847,14 @@ protected:
     };
 
 private:
+    Q_DISABLE_COPY(Codegen)
     VolatileMemoryLocations scanVolatileMemoryLocations(QQmlJS::AST::Node *ast);
     void handleConstruct(const Reference &base, QQmlJS::AST::ArgumentList *args);
     void throwError(ErrorType errorType, const QQmlJS::SourceLocation &loc,
                     const QString &detail);
-    std::optional<Moth::BytecodeGenerator::Label> traverseOptionalChain(QQmlJS::AST::Node *node);
+    bool traverseOptionalChain(QQmlJS::AST::Node *node);
+    void optionalChainFinalizer(const Reference &expressionResult, bool tailOfChain,
+                                bool isDeleteExpression = false);
     Reference loadSubscriptForCall(const Reference &base);
     void generateThrowException(const QString &type, const QString &text = QString());
 };
