@@ -27,17 +27,17 @@
 
 QT_BEGIN_NAMESPACE
 
-class Q_QMLCOMPILER_PRIVATE_EXPORT QQmlJSCodeGenerator : public QQmlJSCompilePass
+class Q_QMLCOMPILER_EXPORT QQmlJSCodeGenerator : public QQmlJSCompilePass
 {
 public:
     QQmlJSCodeGenerator(const QV4::Compiler::Context *compilerContext,
-                       const QV4::Compiler::JSUnitGenerator *unitGenerator,
-                       const QQmlJSTypeResolver *typeResolver,
-                       QQmlJSLogger *logger);
+                        const QV4::Compiler::JSUnitGenerator *unitGenerator,
+                        const QQmlJSTypeResolver *typeResolver, QQmlJSLogger *logger,
+                        BasicBlocks basicBlocks, InstructionAnnotations annotations);
     ~QQmlJSCodeGenerator() = default;
 
-    QQmlJSAotFunction run(const Function *function, const InstructionAnnotations *annotations,
-                          QQmlJS::DiagnosticMessage *error);
+    QQmlJSAotFunction run(const Function *function, QQmlJS::DiagnosticMessage *error,
+                          bool basicBlocksValidationFailed);
 
 protected:
     struct CodegenState : public State
@@ -49,7 +49,7 @@ protected:
     // This is an RAII helper we can use to automatically convert the result of "inflexible"
     // operations to the desired type. For example GetLookup can only retrieve the type of
     // the property we're looking up. If we want to store a different type, we need to convert.
-    struct Q_QMLCOMPILER_PRIVATE_EXPORT AccumulatorConverter
+    struct Q_QMLCOMPILER_EXPORT AccumulatorConverter
     {
         Q_DISABLE_COPY_MOVE(AccumulatorConverter);
         AccumulatorConverter(QQmlJSCodeGenerator *generator);
@@ -133,9 +133,9 @@ protected:
     void generate_PopScriptContext() override;
     void generate_PopContext() override;
     void generate_GetIterator(int iterator) override;
-    void generate_IteratorNext(int value, int done) override;
-    void generate_IteratorNextForYieldStar(int iterator, int object) override;
-    void generate_IteratorClose(int done) override;
+    void generate_IteratorNext(int value, int offset) override;
+    void generate_IteratorNextForYieldStar(int iterator, int object, int offset) override;
+    void generate_IteratorClose() override;
     void generate_DestructureRestElement() override;
     void generate_DeleteProperty(int base, int index) override;
     void generate_DeleteName(int name) override;
@@ -242,11 +242,14 @@ protected:
                              const QQmlJSRegisterContent &to,
                              const QString &variable);
 
-    QString errorReturnValue();
+    void generateReturnError();
     void reject(const QString &thing);
 
     QString metaTypeFromType(const QQmlJSScope::ConstPtr &type) const;
     QString metaTypeFromName(const QQmlJSScope::ConstPtr &type) const;
+    QString compositeMetaType(const QString &elementName) const;
+    QString compositeListMetaType(const QString &elementName) const;
+
     QString contentPointer(const QQmlJSRegisterContent &content, const QString &var);
     QString contentType(const QQmlJSRegisterContent &content, const QString &var);
 
@@ -260,11 +263,14 @@ protected:
     void generateEnumLookup(int index);
 
     QString registerVariable(int index) const;
+    QString lookupVariable(int lookupIndex) const;
     QString consumedRegisterVariable(int index) const;
     QString consumedAccumulatorVariableIn() const;
 
     QString changedRegisterVariable() const;
     QQmlJSRegisterContent registerType(int index) const;
+    QQmlJSRegisterContent lookupType(int lookupIndex) const;
+    bool shouldMoveRegister(int index) const;
 
     QString m_body;
     CodegenState m_state;
@@ -273,7 +279,18 @@ protected:
 
 private:
     void generateExceptionCheck();
-    void generateEqualityOperation(int lhs, const QString &function, bool invert);
+
+    void generateEqualityOperation(
+            const QQmlJSRegisterContent &lhsContent, const QString &lhsName,
+            const QString &function, bool invert) {
+        generateEqualityOperation(
+                lhsContent, m_state.accumulatorIn(), lhsName, m_state.accumulatorVariableIn,
+                function, invert);
+    }
+
+    void generateEqualityOperation(
+            const QQmlJSRegisterContent &lhsContent, const QQmlJSRegisterContent &rhsContent,
+            const QString &lhsName, const QString &rhsName, const QString &function, bool invert);
     void generateCompareOperation(int lhs, const QString &cppOperator);
     void generateArithmeticOperation(int lhs, const QString &cppOperator);
     void generateShiftOperation(int lhs, const QString &cppOperator);
@@ -285,9 +302,16 @@ private:
     void generateInPlaceOperation(const QString &cppOperator);
     void generateMoveOutVar(const QString &outVar);
     void generateTypeLookup(int index);
-    void generateVariantEqualityComparison(const QQmlJSRegisterContent &nonStorable,
-                                           const QString &registerName, bool invert);
+    void generateVariantEqualityComparison(
+            const QQmlJSRegisterContent &nonStorable, const QString &registerName, bool invert);
+    void generateVariantEqualityComparison(
+            const QQmlJSRegisterContent &storableContent, const QString &typedRegisterName,
+            const QString &varRegisterName, bool invert);
+    void generateArrayInitializer(int argc, int argv);
+    void generateWriteBack(int registerIndex);
     void rejectIfNonQObjectOut(const QString &error);
+    void rejectIfBadArray();
+
 
     QString eqIntExpression(int lhsConst);
     QString argumentsList(int argc, int argv, QString *outVar);
@@ -299,22 +323,7 @@ private:
     bool inlineConsoleMethod(const QString &name, int argc, int argv);
     bool inlineArrayMethod(const QString &name, int base, int argc, int argv);
 
-    QQmlJSScope::ConstPtr mathObject() const
-    {
-        using namespace Qt::StringLiterals;
-        return m_typeResolver->jsGlobalObject()->property(u"Math"_s).type();
-    }
-
-    QQmlJSScope::ConstPtr consoleObject() const
-    {
-        using namespace Qt::StringLiterals;
-        return m_typeResolver->jsGlobalObject()->property(u"console"_s).type();
-    }
-
-    QQmlJSScope::ConstPtr arrayPrototype() const
-    {
-        return m_typeResolver->arrayType()->baseType();
-    }
+    void generate_GetLookupHelper(int index);
 
     QString resolveValueTypeContentPointer(
             const QQmlJSScope::ConstPtr &required, const QQmlJSRegisterContent &actual,
@@ -330,7 +339,6 @@ private:
     QHash<int, QString> m_labels;
 
     const QV4::Compiler::Context *m_context = nullptr;
-    const InstructionAnnotations *m_annotations = nullptr;
 
     bool m_skipUntilNextLabel = false;
 
@@ -340,17 +348,20 @@ private:
     {
         QString internalName;
         int registerIndex = -1;
+        int lookupIndex = QQmlJSRegisterContent::InvalidLookupIndex;
 
     private:
         friend size_t qHash(const RegisterVariablesKey &key, size_t seed = 0) noexcept
         {
-            return qHashMulti(seed, key.internalName, key.registerIndex);
+            return qHashMulti(seed, key.internalName, key.registerIndex, key.lookupIndex);
         }
 
         friend bool operator==(
                 const RegisterVariablesKey &lhs, const RegisterVariablesKey &rhs) noexcept
         {
-            return lhs.registerIndex == rhs.registerIndex && lhs.internalName == rhs.internalName;
+            return lhs.registerIndex == rhs.registerIndex
+                    && lhs.lookupIndex == rhs.lookupIndex
+                    && lhs.internalName == rhs.internalName;
         }
 
         friend bool operator!=(
