@@ -17,7 +17,7 @@
 
 #include "qquickwidget.h"
 #include <private/qwidget_p.h>
-#include <private/qrhi_p.h>
+#include <rhi/qrhi.h>
 #include <private/qbackingstorerhisupport_p.h>
 
 #include <QtCore/qurl.h>
@@ -67,7 +67,7 @@ public:
     QPlatformTextureList::Flags textureListFlags() override;
     QImage grabFramebuffer() override;
 
-    void init(QQmlEngine* e = 0);
+    void init(QQmlEngine* e = nullptr);
     void ensureBackingScene();
     void initOffscreenWindow();
     void ensureEngine() const;
